from setuptools import setup

setup(
    name='RbxAPI',
    version='0.21.0',
    description="RbxAPI is a Pythonic and efficient Roblox API wrapper. With many quality of life adjustments made during its first iterations, it aims to give you the simplest way of accessing Roblox's API endpoints!",
    author='PythonicIconic',
    author_email='coatedalmond@protonmail.com',
    packages=['RbxAPI'],
    install_requires=['requests']
)